; KbqLib.asm: version 1.01

	include "doorsos.h"	;This line can be changed to the library that
				; you prefer.

	xdef _library
	xdef _ti89
	xdef _ti92plus

	xdef	kbqlib@0000
	xdef	kbqlib@0001
	xdef	kbqlib@0002
	xdef	kbqlib@0003
	xdef	kbqlib@0004

;----------------------------------------------------------------------------------------
; Introduction:
;
;	Hello!  Welcome to the kbqlib library.  This library was designed by Michael
; Olson between 9/2000 and 11/2000.
;
; Warranty (and guidelines):
;
;	This program is released under the GNU Lesser General Public License.
;	See license.txt for more info.  If you did not get this file, please go to
;	http://www.gnu.org and download it.
;
;	Feel free to use these procedures and methods.  It would be nice
;	 if you could mention me in your THANKS section, but I do not require it.
;
;	If you want to make changes to this work, go ahead!  If you have time, it
;	 would be great if you could send me your improvements so I can take a
;	 look at them.
;
; Goals:
;
;   The goals for this project are:
;
;	-to make a set of keyboard routines that are not only compatible with every 68k 
;	 calculator yet made by TI, but are also (hopefully) compatible with future 
;	 AMS versions!
;
;	-to make it very easy to substitute outdated keyboard handling routines with new
;	 ones that will work properly.
;
;	-to have this library be compatible with TeOS, DoorsOS, and Universal OS. This
;	 is why I have referred to _ROM_CALL_xxx instead of something like:
;	 	tios::ngetchx	or	doorsos::GKeyFlush.
;	 That way, all you have to do is change the "include" statement at the top of
;	 this file to whichever file you want to use.
;
;	-to make it easy to update the routines that the modified programs will use.
;	 This is why I have chosen to make this an on-calculator library instead of
;	 my original intent (a compile-time library).  This also saves space, hopefully.
;
;	-to make a direct substitute for kb_vars. (But I am not good enough as a
;	 programmer.)  For example, being able to substitute:
;		tst.w (doorsos::kb_vars+1c)
;	 with
;		tst.w (kbqlib::kb_vars+1c).
;	 Unfortunately, this easy method did not seem to work out!
;
; Directions (Non-developer):
;	# Please see ..\readme.txt!
;
; Directions (Assembly programming):
;
;	# Read the source code here.
;	# Copy the kbqlib.h file to your include directory.
;	# Add this statement to the beginning of the program you would like to convert:
;		include "kbqlib.h"
;	# Add the necessary statements to substitute old commands for the new
;	  functions.
;	# Compile and link your program.
;	# If you get an error, make sure that kbqlib.h is in the include path of
;	  your compiler!
;
; Directions (C programming):
;
;	# Copy kbqlib.hc to your include directory.
;	# Add this statement to the beginning of your program:
;		include "kbqlib.hc" /* or include <kbqlib.hc> */
;	# Add the necessary substitutions to your programs.
;	# Compile and link your program.
;
; ASM Usage examples:
;
;	Here are some examples:
;
;	Change:
;		move.w (doorsos::kb_vars+1e),d0
;		tst.w d0
;		bne KeyStrokeEntered
;	to:
;		jsr kbqlib::kbqhit
;		tst.w d0
;		bne KeyStrokeEntered
;
;	and change:
;		clr.w (tios::kb_vars+1c)
;	to:
;		jsr kbqlib::kbqclr
;
;	See the included programs for the best examples... ;)
;
; C Usage example:
;
;	include "kbqlib.hc"	// you might need to use:
;				// 	include <kblib.hc>
;				// intead of this command
;
;	int keyNeeded;		// Var that records a keystroke
;
;	keyNeeded=kbqget();	// to put the key code in the keyNeeded var
;
;	printf(keyNeeded);	// print the key code on screen
;
;   For a better example, see the LINES directory...
;
;
; All of the Functions at a Glance:
;
; Main:
;	kbd_queue(0000)	- The routine used internally to access the keyboard buffer
;			  (also called a queue).  WARNING: Only use this if you have
;			  experimented with it and know what you are doing!
;	kbqclr(0001)	- Clears the keyboard buffer.
;	kbqget(0002)	- Get a keystroke, taking it out if there is one.
;	kbqhit(0003)	- Return the value of a key, if pressed.
;	kbqput(0004)	- Put a keystroke into the buffer.
;
; Alternates:
;	normal_clr	- kbqclr routine for the calcs that have the needed ROM CALLS
;	normal_get	- kbqget routine for the calcs that have the needed ROM CALLS
;	normal_hit	- kbqhit routine for the calcs that have the needed ROM CALLS
;	normal_put	- kbqput routine for the calcs that have the needed ROM CALLS
;
; Internal:
;	finish_it	- Free the memory that we used and return d0.
;	clr_loop	- A loop needed to help the kbqclr function.
;
; Variables:
;	tmp1		- Temporary longword var needed as an address to store the
;			  results of some of the ROM_CALLS.
;
; How Did I Do It?
;
;	I came across the kbq_queue routine when I was looking at the TIGCCLIB
; documentation.  I realized that this could be the solution to the major
; compatibility problem with all new AMS versions: the relocation of the kb_vars!
; I made a sample program in TI-GCC to see how this kbd_queue thing worked.  The sample
; ran correctly on my HW2 AMS 2.05 calculator, so I disassembled it using TI92DASM to
; get the code.
;	Then, I made a bigger example program (included) called kbqlib.c and
; disassembled it.  This one included all of the main functions.
; 	I tweaked the ASM code further to add TI-92+ AMS 1.00 support.  It checks to see
; if the calculator supports the needed _ROM_CALL.  If it does, run the normal routine.
; Otherwise, run the special routine. I also had to make sure that the stack was returned
; to its proper state as I had problems with the link statements. (In other words, I did 
; not understand them.  Thanks Zeljko Juric for trying to educate me!) So, I took out all
; of the "link" statements because many programmers prefer to send their parameters and 
; get the results through d0-d1.  I used the bss variable "tmp1", then, to store the
; resulting keystrokes, and d0 contains the final result of each function.
;
; Footnotes:
;	* Look at the warning in the code to kbd_queue routine!
;	* feel free to e-mail me if something goes awry.
;
; Credits:
;
;  see ..\readme.txt
;
;
; Contacting me:
; 	e-mail:
;	Michael Olson at <bigmike160@yahoo.com>
;----------------------------------------------------------------------------------------

kbqlib@0000:
kbd_queue:
	moveq.l	#$6,d0
	trap	#$9
	rts
;
; If you look at interrupts.txt that came with DoorsOS, you might notice that there
;  is something interesting listed under INT 9.  If you set d0.w=6 and trap #9, you
;  will receive an address in the a0.l register.  This address SEEMS to be a direct link
;  to the keyboard buffer.  If you read from it (by using the queue manipulation
;  _ROM_CALL's), you can get a keycode, and if you write a code to it, you will put a
;  keystroke into the buffer.
; I have tried to use kbq_queue as a direct substitute for the kb_vars without success.
;  Unless you know what you are doing (and I don't), it is better to use my included 
;  subroutines.
;
;----------------------------------------------------------------------------------------
kbqlib@0003:
kbqhit:
; Get a keystroke but don't take it out of the buffer
; Input:	nothing
; Output:	d0.w=key code or
;		d0.w=0 if no key was pressed
;----------------------------------------------------------------------------------------
	movem.l	d1-d7/a0-a6,-(a7)	;Save our precious registers.
	movea.l	$C8,a0		;Some address having to do with _ROM_CALLs ...
	cmpi.l	#$3A8,-4(a0)	;Do we have at least $3ab = 939 entries?
	bhi.s	normal_hit	;Yes, go to the BEST routine.
				;Otherwise, use an alternate compatible routine.
	jsr	_ROM_CALL_052	;kbhit call: find out if there is a keystroke in buffer.
	move.w	d0,(tmp1)	;Put the result in tmp1 so that it can be returned.
	bra.s	finish_it	;Finish it up.

;----------------------------------------------------------------------------------------
normal_hit:
; Alternative routine
; Used if our AMS version supports ROM_CALLS above $3a8.  This is the "best" routine.
;----------------------------------------------------------------------------------------
	jsr	kbd_queue(pc)	;Retrieve a0=addr to keyboard queue (buffer).
	move.l	a0,-(a7)	;Push it first.
	pea	tmp1		;Then the address where the results will be stored.
	jsr	_ROM_CALL_3AB	;OSqinquire call: Does there exist an element in the
				; queue? If so, store it in tmp1, our first parameter
				; (based on C standards).
	addq.l	#8,a7		;RETRIEVE our stack space.

;----------------------------------------------------------------------------------------
finish_it:
; Internal routine
; Returns the correct value for d0 and restores the stack to its correct place.
;----------------------------------------------------------------------------------------
	move.w	(tmp1),d0	;Return our results to d0
	bclr	#11,d0		;Enable autorepeat (see kbd_queue explanation
				; in the tigcclib docs)
	movem.l	(a7)+,d1-d7/a0-a6	;restore the saved registers
	rts

;----------------------------------------------------------------------------------------
kbqlib@0001:
kbqclr:
; Clear the keyboard buffer (or queue, whichever term you like better)
; Input:	nothing
; Output:	d0.w=0 always
;----------------------------------------------------------------------------------------
	movem.l	d1-d7/a0-a6,-(a7)
	movea.l	$C8,a0
	cmpi.l	#$3A8,-4(a0)
	bhi.s	normal_clr
	jsr	_ROM_CALL_180	;GKeyFlush call: Clears the kb buffer.
	move.w	#0,(tmp1)
	bra.s	finish_it

;----------------------------------------------------------------------------------------
normal_clr:
; Alternative routine
; Used if our AMS version supports ROM_CALLS above $3a8.  This is the "best" routine.
;----------------------------------------------------------------------------------------
	jsr	kbd_queue(pc)

clr_loop:
	bsr	kbqget		;OSqclear did not seem to work correctly, so I used
	tst.w	d0		; this little clearing loop instead, which calls
	bne.s	clr_loop	; OSdequeue until all keystrokes are out.

	move.w	#0,(tmp1)
	bra.s	finish_it

;----------------------------------------------------------------------------------------
kbqlib@0002:
kbqget:
; Get a key from the buffer without replacement (takes it out)
; Input:	nothing
; Output:	d0.w=key code
;----------------------------------------------------------------------------------------
	movem.l	d1-d7/a0-a6,-(a7)
	movea.l	$C8,a0
	cmpi.l	#$3A8,-4(a0)
	bhi.s	normal_get
	jsr	_ROM_CALL_052	;kbhit call: First check to see if we have a key there.
				; If you call ngetchx, the calc will keep waiting for
				; a key, which is NOT what we want!
	move.w	d0,(tmp1)
	tst.w	d0		;Do we have a key?
	beq.s	finish_it	;No, go no further.
	jsr	_ROM_CALL_051	;ngetchx call: Wait until a key is pressed and then
				; take it out of the buffer.
	move.w	d0,(tmp1)
	bra.s	finish_it

;----------------------------------------------------------------------------------------
normal_get:
; Alternative routine
; Used if our AMS version supports ROM_CALLS above $3a8.  This is the "best" routine.
;----------------------------------------------------------------------------------------
	jsr	kbd_queue(pc)
	move.l	a0,-(a7)
	pea	tmp1
	jsr	_ROM_CALL_3AA	;OSdequeue call: If a queue contains an element then
				; take it out and return 0. Otherwise return 1.
	addq.l	#8,a7
	tst.w	d0
	beq	finish_it
	clr.w	(tmp1)		;Force tmp1 to equal 0. Otherwise it will still record
				; a keystroke (very bad).
	bra	finish_it

;----------------------------------------------------------------------------------------
kbqlib@0004:
kbqput:
; Put a key code into the buffer (remember that it is First In First Out)
; Input:	d0.w=key code to be pushed into keystroke buffer
; Output:	d0.w=key code if successful, otherwise
;		d0.w=0 if there is an error (buffer is full).
;----------------------------------------------------------------------------------------
	movem.l	d1-d7/a0-a6,-(a7)
	movea.l	$C8,a0
	cmpi.l	#$3A8,-4(a0)
	bhi.s	normal_put
	jsr	_ROM_CALL_050	;pushkey call: Put a key into the buffer, remove one if
				; if it is already there.
	move.w	#0,(tmp1)
	bra	finish_it

;----------------------------------------------------------------------------------------
normal_put:
; Alternative routine
; Used if our AMS version supports ROM_CALLS above $3a8.  This is the "best" routine.
;----------------------------------------------------------------------------------------
	jsr	kbd_queue(pc)
	clr.w	(tmp1)		;make sure that it is cleared initially.
	move.l	a0,-(a7)
	move.w	d0,-(a7)
	jsr	_ROM_CALL_3A9	;OSenqueue: Put an element into a queue.
	move.w	(a7)+,d1	;Get the initial key code back.
	addq.l	#4,a7
	tst.w	d0
	beq	finish_it	;unsuccessful: d0.w=0
	move.w	d1,(tmp1)	;successful:d0.w=initial code
	bra	finish_it

;----------------------------------------------------------------------------------------
; Variables
;----------------------------------------------------------------------------------------
	BSS
tmp1	ds.l	1		;Used as a memory location that stores the result.
				; This form is nice because we can treat it as an
				; address register.

;----------------------------------------------------------------------------------------
	end
